import pytest

from nuclia.lib.nua_responses import PushPayload


def test_uuid_validation():
    PushPayload()
    with pytest.raises(ValueError):
        PushPayload(kbid="")
    with pytest.raises(ValueError):
        PushPayload(uuid="")
    with pytest.raises(ValueError):
        PushPayload(uuid="invalid.uuid")



import json

from datetime import datetime, timezone
from base64 import b64decode

try:
    from litellm import CustomLLM
    from litellm.llms.custom_httpx.http_handler import HTTPHandler
    from litellm.utils import ModelResponse, Choices, Message
except ImportError:
    raise ImportError(
        "The 'litellm' library is required to use this functionality. "
        "Install it with: pip install nuclia[litellm]"
    )

from nuclia.lib.nua import NuaClient
from nuclia.sdk.predict import NucliaPredict
from nuclia.lib.nua_responses import ChatModel, UserPrompt
from nuclia_models.predict.generative_responses import (
    GenerativeFullResponse,
)
from typing import Callable, Optional, Union

import httpx


class NucliaNuaChat(CustomLLM):
    """
    A LiteLLM custom Model that uses nua client under the hood
    """

    def __init__(self, token: str):
        self.token = token
        if self.token:
            regional_url, expiration_date = self._parse_token(self.token)
            now = datetime.now(timezone.utc)
            if expiration_date <= now:
                raise ValueError("Expired nua token")
            self.region_base_url = regional_url

        self.nc_sync = NuaClient(
            region=self.region_base_url,
            token=self.token,
            account="",  # Not needed for current implementation, required by the client
        )
        self.predict_sync = NucliaPredict()

    @staticmethod
    def _parse_token(token: str):
        parts = token.split(".")
        if len(parts) < 3:
            raise ValueError("Invalid JWT token, missing segments")

        b64_payload = parts[1]
        payload = json.loads(b64decode(b64_payload + "=="))
        regional_url = payload["iss"]
        token_expire_ts = payload["exp"]
        if token_expire_ts >= 32536850400:
            # sc-11523
            token_expire_ts = 32536850399

        expiration_date = datetime.fromtimestamp(token_expire_ts, tz=timezone.utc)
        return regional_url, expiration_date

    def _process_messages(self, messages: list[dict[str, str]]) -> tuple[str, str]:
        system_messages = []
        user_messages = []

        for msg in messages:
            role = msg["role"]
            content = msg["content"].strip()

            if role == "system":
                system_messages.append(content)
            else:
                user_messages.append(content)

        formatted_system = "\n".join(system_messages) if system_messages else ""
        formatted_user = "\n".join(user_messages) if user_messages else ""

        return formatted_system, formatted_user

    def completion(
        self,
        model: str,
        messages: list,
        api_base: str,
        custom_prompt_dict: dict,
        model_response: ModelResponse,
        print_verbose: Callable,
        encoding,
        api_key,
        logging_obj,
        optional_params: dict,
        acompletion=None,
        litellm_params=None,
        logger_fn=None,
        headers={},
        timeout: Optional[Union[float, httpx.Timeout]] = None,
        client: Optional[HTTPHandler] = None,
    ) -> ModelResponse:
        if not self.predict_sync or not self.nc_sync:
            raise RuntimeError("Sync clients not initialized.")

        system_prompt, user_prompt = self._process_messages(messages)
        body = ChatModel(
            question="",
            retrieval=False,
            user_id="nua-chat",
            system=system_prompt,
            user_prompt=UserPrompt(prompt=user_prompt),
            query_context={},
            format_prompt=False,
        )
        response: GenerativeFullResponse = self.predict_sync.generate(
            text=body,
            model=model,
            nc=self.nc_sync,
        )

        return ModelResponse(
            choices=[Choices(message=Message(content=response.answer))]
        )



from nuclia.sdk.kb import NucliaKB, AsyncNucliaKB
from nuclia.tests.fixtures import IS_PROD
from nuclia_models.worker.tasks import (
    TaskList,
    TaskName,
    ApplyOptions,
    TaskResponse,
    JobStatus,
)
from nuclia_models.worker.proto import (
    DataAugmentation,
    ApplyTo,
    Filter,
    Operation,
    LLMConfig,
)
import pytest


def test_worker_manager_tasks(testing_config):
    if not IS_PROD:
        assert True
        return

    nkb = NucliaKB()

    output = nkb.task.list()
    assert isinstance(output, TaskList)

    output = nkb.task.start(
        task_name=TaskName.LABELER,
        apply=ApplyOptions.EXISTING,
        parameters=DataAugmentation(
            name="test",
            on=ApplyTo.FIELD,
            filter=Filter(),
            operations=[Operation()],
            llm=LLMConfig(),
        ),
    )
    assert isinstance(output, TaskResponse)
    task_id = output.id

    output = nkb.task.get(task_id=task_id)
    assert output.request.id == task_id

    output = nkb.task.stop(task_id=task_id)
    assert output.id == task_id
    assert output.status == JobStatus.NOT_RUNNING

    output = nkb.task.restart(task_id=task_id)
    assert output.id == task_id
    assert output.status == JobStatus.NOT_RUNNING

    output = nkb.task.stop(task_id=task_id)
    assert output.id == task_id
    assert output.status == JobStatus.NOT_RUNNING

    output = nkb.task.delete(task_id=task_id)
    assert output is None

    output = nkb.task.get(task_id=task_id)
    assert output.request is None
    assert output.config is None

    # Delete all tasks
    output = nkb.task.list()
    assert isinstance(output, TaskList)
    for task in output.running:
        nkb.task.stop(task_id=task.id)
        nkb.task.delete(task_id=task.id)
    for task in output.done:
        nkb.task.delete(task_id=task.id)


@pytest.mark.asyncio
async def test_worker_manager_tasks_async(testing_config):
    if not IS_PROD:
        assert True
        return

    nkb = AsyncNucliaKB()

    output = await nkb.task.list()
    assert isinstance(output, TaskList)

    output = await nkb.task.start(
        task_name=TaskName.LABELER,
        apply=ApplyOptions.EXISTING,
        parameters=DataAugmentation(
            name="test",
            on=ApplyTo.FIELD,
            filter=Filter(),
            operations=[Operation()],
            llm=LLMConfig(),
        ),
    )
    assert isinstance(output, TaskResponse)
    task_id = output.id

    output = await nkb.task.get(task_id=task_id)
    assert output.request.id == task_id

    output = await nkb.task.stop(task_id=task_id)
    assert output.id == task_id
    assert output.status == JobStatus.NOT_RUNNING

    output = await nkb.task.restart(task_id=task_id)
    assert output.id == task_id
    assert output.status == JobStatus.NOT_RUNNING

    output = await nkb.task.stop(task_id=task_id)
    assert output.id == task_id
    assert output.status == JobStatus.NOT_RUNNING

    output = await nkb.task.delete(task_id=task_id)
    assert output is None

    output = await nkb.task.get(task_id=task_id)
    assert output.request is None
    assert output.config is None

    # Delete all tasks
    output = await nkb.task.list()
    assert isinstance(output, TaskList)
    for task in output.running:
        await nkb.task.stop(task_id=task.id)
        await nkb.task.delete(task_id=task.id)
    for task in output.done:
        await nkb.task.delete(task_id=task.id)



from nuclia_models.predict.generative_responses import TextGenerativeResponse

from nuclia.lib.nua_responses import ChatModel, UserPrompt
from nuclia.sdk.predict import AsyncNucliaPredict, NucliaPredict
import pytest
from nuclia_models.predict.remi import RemiRequest


def test_predict(testing_config):
    np = NucliaPredict()
    embed = np.sentence(text="This is my text", model="multilingual-2024-05-06")
    assert embed.time > 0
    assert len(embed.data) == 1024


def test_predict_query(testing_config):
    np = NucliaPredict()
    query = np.query(
        text="Ramon, this is my text",
        semantic_model="multilingual-2024-05-06",
        token_model="multilingual",
        generative_model="chatgpt-azure-4o-mini",
    )
    assert query.language == "en"
    assert query.visual_llm is True
    assert query.entities and query.entities.tokens[0].text == "Ramon"
    assert query.sentence and len(query.sentence.data) == 1024


def test_rag(testing_config):
    np = NucliaPredict()
    generated = np.rag(
        question="Which is the CEO of Nuclia?",
        context=[
            "Nuclia CTO is Ramon Navarro",
            "Eudald Camprubí is CEO at the same company as Ramon Navarro",
        ],
        model="chatgpt-azure-4o-mini",
    )
    assert "Eudald" in generated.answer


def test_generative(testing_config):
    np = NucliaPredict()
    generated = np.generate(text="How much is 2 + 2?", model="chatgpt-azure-4o-mini")
    assert "4" in generated.answer


@pytest.mark.asyncio
async def test_async_generative(testing_config):
    np = AsyncNucliaPredict()
    generated = await np.generate(
        text="How much is 2 + 2?", model="chatgpt-azure-4o-mini"
    )
    assert "4" in generated.answer


def test_stream_generative(testing_config):
    np = NucliaPredict()
    found = False
    for stream in np.generate_stream(
        text="How much is 2 + 2?", model="chatgpt-azure-4o-mini"
    ):
        if isinstance(stream.chunk, TextGenerativeResponse) and stream.chunk.text:
            if "4" in stream.chunk.text:
                found = True
    assert found


@pytest.mark.asyncio
async def test_async_stream_generative(testing_config):
    np = AsyncNucliaPredict()
    async for stream in np.generate_stream(
        text="How much is 2 + 2?", model="chatgpt-azure-4o-mini"
    ):
        if isinstance(stream.chunk, TextGenerativeResponse) and stream.chunk.text:
            if "4" in stream.chunk.text:
                found = True
    assert found


SCHEMA = {
    "name": "ClassificationReverse",
    "description": "Correctly extracted with all the required parameters with correct types",
    "parameters": {
        "$defs": {
            "Options": {
                "enum": ["SPORTS", "POLITICAL"],
                "title": "Options",
                "type": "string",
            }
        },
        "properties": {
            "title": {"default": "label", "title": "Title", "type": "string"},
            "description": {
                "default": "Define labels to classify the subject of the document",
                "title": "Description",
                "type": "string",
            },
            "document_type": {
                "description": "Type of document, SPORT example: elections, Illa, POLITICAL example: football",
                "items": {"$ref": "#/$defs/Options"},
                "title": "Document Type",
                "type": "array",
            },
        },
        "required": ["document_type"],
        "type": "object",
    },
}

TEXT = """"Many football players have existed. Messi is by far the greatest. Messi was born in Rosario, 24th of June 1987"""


@pytest.mark.asyncio
async def test_nua_parse(testing_config):
    np = AsyncNucliaPredict()
    results = await np.generate(
        text=ChatModel(
            question="",
            retrieval=False,
            user_id="Nuclia PY CLI",
            user_prompt=UserPrompt(prompt=TEXT),
            json_schema=SCHEMA,
        )
    )
    assert "SPORTS" in results.object["document_type"]


def test_nua_remi(testing_config):
    np = NucliaPredict()
    results = np.remi(
        RemiRequest(
            user_id="Nuclia PY CLI",
            question="What is the capital of France?",
            answer="Paris is the capital of france!",
            contexts=[
                "Paris is the capital of France.",
                "Berlin is the capital of Germany.",
            ],
        )
    )
    assert results.answer_relevance.score >= 4

    assert results.context_relevance[0] >= 4
    assert results.groundedness[0] >= 4

    assert results.context_relevance[1] < 2
    assert results.groundedness[1] < 2


@pytest.mark.asyncio
async def test_nua_async_remi(testing_config):
    np = AsyncNucliaPredict()
    results = await np.remi(
        RemiRequest(
            user_id="Nuclia PY CLI",
            question="What is the capital of France?",
            answer="Paris is the capital of france!",
            contexts=[
                "Paris is the capital of France.",
                "Berlin is the capital of Germany.",
            ],
        )
    )
    assert results.answer_relevance.score >= 4

    assert results.context_relevance[0] >= 4
    assert results.groundedness[0] >= 4

    assert results.context_relevance[1] < 2
    assert results.groundedness[1] < 2



from nuclia.sdk.kb import NucliaKB
from nucliadb_sdk.v2.exceptions import NotFoundError

from time import sleep


def test_graph(testing_config):
    nkb = NucliaKB()
    nkb.add_graph(
        slug="graph1",
        graph=[
            {
                "source": {"group": "People", "value": "Alice"},
                "destination": {"group": "People", "value": "Bob"},
                "label": "is friend of",
            },
            {
                "source": {"group": "People", "value": "Alice"},
                "destination": {"group": "City", "value": "Toulouse"},
                "label": "lives in",
            },
            {
                "source": {"group": "People", "value": "Bob"},
                "destination": {"group": "Food", "value": "Cheese"},
                "label": "eat",
            },
        ],
    )
    relations = nkb.get_graph(slug="graph1")
    assert len(relations) == 3
    nkb.update_graph(
        slug="graph1",
        graph=[
            {
                "source": {"group": "People", "value": "Victor"},
                "destination": {"group": "Food", "value": "Cheese"},
                "label": "eat",
            },
        ],
    )
    relations = nkb.get_graph(slug="graph1")
    assert len(relations) == 4
    nkb.delete_graph(slug="graph1")
    try:
        sleep(0.5)
        nkb.get_graph(slug="graph1")
        assert False
    except NotFoundError:
        assert True
